"""Utilities for extended state machines.

get_lts_from_dotfile(dot_file) - Get label transition system from graphviz dot file
print_lts(start_state,transitions) - Print labeled transition system

parse_labels_of_lts(transitions)- Parse labels of labeled transition system
read_abstractions_from_jsonfile(json_filename) - Read data from learner
apply_mapper_data_on_transitions(..) - Apply data from mapper on learner results

print_transitions(transitions) - Print transitions
print_transitions_long(transitions) - Print transitions in long format
get_locations(transitions) - Get set of all locations in statemachine
split_input_and_output(transitions) - Split transitions with both an input
                                      and output event into two transitions
get_uppaal_model() - Get uppaal model


"""

# Author: Harco Kuppens

__all__ = ["get_lts_from_dotfile",
           "print_lts",
           "parse_labels_of_lts",
           "read_abstractions_from_jsonfile",
           "apply_mapper_data_on_transitions",
           "print_transitions",
           "print_transitions_long",
           "get_locations",
           "split_input_and_output",
           "get_uppaal_model",
          ]

import sys, re, json, copy # module from standard lib (python 2.6 and later)

import pyuppaal
#import pygraphviz


#-------------------------------------------------------------------------------
#  pre processing
#-------------------------------------------------------------------------------



def get_lts_from_dotfile(dot_file):
    """ Get labeled transition system from graphviz dot file

        The dot file:
           - describes a digraph with labels
           - encodes the start state with the color='red' attribute
             note: this corresponds with the highlighted state in learnlib API

        Returns: [start_state,transions]
        Where :
           - start_state: start state label
           - transitions: list of transitions

    """

    start_state='unknown'
    f=file(dot_file)
    lines=f.readlines()

    # find start state
    for line in lines:
        line=line.replace('label=""','')
        if line.find('[color="red"]') != -1:
            start_state=line[:line.find(' ')]
            break


    # get transitions
    transitions=[]
    for line in lines:
        if line.find('->') != -1:
            transitions.append(line)

    # throw away transitions with the keywords : quiescence or inconsistency or undefined
    transitions = [ t for t in transitions if ( 'quiescence' not in t ) and ( 'inconsistency' not in t  )  and ( 'undefined' not in t  )]

    trans_out=[]
    regexpr_transition=re.compile(r'\s*(\w*)\s*-\>\s*(\w*)\s*\[label=\<(.*)\>\]')
    regexpr_tag=re.compile(r'<[^>]+>')
    for transition in transitions:
        match=regexpr_transition.match(transition)
        if match:
            match=match.groups()
            label=regexpr_tag.sub('',match[2])
            trans_out.append({
              'source' : match[0],
              'target' : match[1],
              'label': label
            })

    return [start_state,trans_out]

#     trans_out=[]
#     start_state='unknown'
#     G=pygraphviz.AGraph(dot_file)
#     transitions=G.edges()
#     for t in transitions:
#         label=t.attr['label']
#         label=re.sub('<[^<]+?>', '', label)
#         trans_out.append({
#           'source' : str(t[0]),
#           'target' : str(t[1]),
#           'label': label
#         })
#     for n in G.nodes():
#         if n.attr['color'] == 'red':
#             start_state=n.attr['label']
#     return [start_state,trans_out]

def print_lts(start_state,transitions):
    """Print labeled transition system

       The transitions are printed as used in a statemachine::

          source -> target : label

    """
    indent='  '
    print "start state: " + start_state
    for t in transitions:
        print t['source'] + ' -> ' + t['target']  + ' : ' + t['label']


def parse_labels_of_lts(transitions):
    """Parse labels of labeled transition system
    """
    trans_out=[]
    for t in transitions:
        label=t['label']
        [inputstr,outputstr]=label.split('/')
        input=inputstr.split('_')
        inputlabel=input[0]
        inputargs=input[1:]
        output=outputstr.split('_')
        outputlabel=output[0]
        outputargs=output[1:]
        trans_out.append({
          'source' : t['source'],
          'target' : t['target'],
          'input': {
              'name': inputlabel,
              'args': inputargs
          },
          'output': {
              'name': outputlabel,
              'args': outputargs
          },
          #'guard':'',
          #'action':''
        })
    return trans_out




def read_abstractions_from_jsonfile(json_filename_in):
    """ Read data from learner

        Reads in json file which contains data from learned model
        learned with learner.
    """
    f=open(json_filename_in,'r')
    mapper=json.load( f )
    f.close()
    #print json.dumps(jsonstr, sort_keys=True, indent=4)


    # output abstactions
    outputs={}
    outs=mapper['outputs']
    for i in outs:
        outputs[ i['name'] ] = i['params']
 
    # input abstactions
    inputs={}
    ins=mapper['inputs']
    for i in ins:
        inputs[ i['name'] ] = i['params']
 
 
    # sut interfaces
    interfaces=mapper['concrete_alphabet']


    cons=mapper['constants']

    return  [inputs,outputs,cons,interfaces]



#-------------------------------------------------------------------------------
#  processing
#-------------------------------------------------------------------------------



# abstraction in  firstlast mode
#     {
#         "methodname": "IRegister", 
#         "paramindex": 1, 
#         "stateid": 0
#     }
# 
def get_guard_from_input(rule,operator,cons,pos,inputlabel):
    if rule['methodname'] == '_CONSTANTS_':
        return "p" + str(pos) + ' ' + operator + ' ' +  str( cons[rule['paramindex']] )
    else:
        if  ( rule['methodname'] == inputlabel ) and ( rule['paramindex'] < pos ) :
            return "p" + str(pos) + ' ' + operator + ' ' + 'p' +  str(rule['paramindex'])
        else:
            return "p" + str(pos) + ' ' + operator + ' ' + rule['occurence'] + '_' + rule['methodname']  + '_p' +  str(rule['paramindex'])

def get_output_param(rule,cons):
    if rule['methodname'] == '_CONSTANTS_':
        return str( cons[rule['paramindex']] )
    else:
        return rule['occurence'] + '_' + rule['methodname']  + '_p' +  str(rule['paramindex'])


def get_input_event(inputlabel,num_input_params):
        # construct input event and its arguments for transition
        # -> constructed from : inputlabel and num_input_params
        inputparams=[]
        for num in xrange(num_input_params):
            inputparams.append('p'+str(num))
        return inputlabel + '(' + ','.join(inputparams)  + ')'

def get_guard(inputlabel,inputargs,cons,input_abstractions):
        # construct guard for transition
         # -> constructed from :  inputlabel,inputargs,cons  and input_abstractions(rules)
        # 1. guards are retreived by applying input_rules on input
        pos=0
        params=[]
        for i in inputargs:
            if int(i) != -1:
                rule=input_abstractions[ inputlabel ][pos][int(i)]
                params.append( get_guard_from_input(rule,'==',cons,pos,inputlabel) )
            else:
                # all rules for inputarg gives != guard
                for rule in input_abstractions[ inputlabel ][pos]:
                    params.append( get_guard_from_input(rule,'!=',cons,pos, inputlabel ) )
            pos=pos+1
        # 2. store collect guard conditions joined with boolean AND in transition
        if params:
            return ' && '.join(params)
        else:
            return '' # no guard

def get_output_event(outputlabel,outputargs,cons,outputs):
        # construct output event and its arguments for transition
        # -> constructed from : outputlabel,outputargs,cons and outputs(rules)
        pos=0
        params=[]
        for i in outputargs:
            if int(i) != -1:
                rule=outputs[ outputlabel ][pos][int(i)]
                params.append( get_output_param(rule,cons) )
            else:
                print "problem output  -1 !!!"
            pos=pos+1

        return outputlabel + '(' + ','.join(params) + ')'


def apply_mapper_data_on_transitions(transitions,inputs,outputs,cons,number_params):
    """ Apply data from mapper on results from learner




    """

    trans_out=[]
    indent='  '
    for t in transitions:
        transition={}
        transition['source']=t['source']
        transition['target']=t['target']


        inputlabel=t['input']['name']
        inputargs=t['input']['args']
        outputlabel=t['output']['name']
        outputargs=t['output']['args']

        num_input_params=number_params[inputlabel]



        transition['output']=get_output_event(outputlabel,outputargs,cons,outputs)
        transition['input']=get_input_event(inputlabel,num_input_params)
        transition['guard']=get_guard(inputlabel,inputargs,cons,inputs)

        # create actions for transition
        #  - when no input event params :  construct single transition with no actions
        #  - with input event params : create two transitions with different guards and actions
        #                              -> one two initialize and one to update
        if num_input_params > 0 :
            # generate two transactions with different guard and action
            transition_dupl=copy.deepcopy(transition)

            # first transition   : initialization statemachine params corresponding with input event's params
            actions=[]
            for num in xrange(num_input_params):
                actions.append('first_' + inputlabel + '_p' + str(num) + ' = p' + str(num) + ';')
                actions.append('last_' + inputlabel + '_p' + str(num) + ' = p' + str(num) + ';')

            transition['action']=' '.join(actions)

            guard='first_' +   inputlabel + '_p0 == -1'
            if transition['guard']:
                transition['guard']+=" && " + guard
            else:
                transition['guard']=guard

            # second transition  : update statemachine params corresponding with input event's params
            actions=[]
            for num in xrange(num_input_params):
                actions.append('last_' + inputlabel + '_p' + str(num) + ' = p' + str(num) + ';')
            transition_dupl['action'] = ' '.join(actions)

            guard='first_' +   inputlabel + '_p0 != -1'
            if transition_dupl['guard']:
                transition_dupl['guard']+=" && " + guard
            else:
                transition_dupl['guard']=guard

            # add transitions to output
            trans_out.append(transition)
            trans_out.append(transition_dupl)
        else :
            transition['action']=''  # no action  and doesn't add guard
            trans_out.append(transition)


    return trans_out


#-------------------------------------------------------------------------------
# post processing
#-------------------------------------------------------------------------------



def print_transitions_long(transitions):
    """Print transitions in long format

       For each transition the guard,input,output,action are printed
       on a separate line::

          <source> -> <target>
             guard: <guard>
             input: <input>
             output: <output>
             action: <action>

    """
    indent='  '
    for t in transitions:
        print t['source'] + ' -> ' + t['target']
        print indent*2, 'guard:',  t['guard']
        #print indent*2, 'guard:', '[ ', t['guard'] , ' ]'
        print indent*2, 'input:', t['input']
        print indent*2, 'output:', t['output']
        print indent*2, 'action:', t['action']

def print_transitions(transitions):
    """Print transitions

       The transitions are printed as used in a statemachine::

          source -> target : [guard] input/output ; action

    """
    indent='  '
    for t in transitions:
        guard=''
        if  t['guard']:
            guard= " [ " + t['guard']  + ' ] '
        action=''
        if  t['action']:
            action = " ; " + t['action']
        print t['source'] + ' -> ' + t['target']  + ' : ' + guard +  t['input'] + ' / ' +  t['output'] +  action
 
def split_input_and_output(transitions):
    """Split transitions with both an input and output event into two transitions

       Makes two sequential transitions with a dummy state in between:
        - dummy state <midstate> is labeled :
             m_<counter>
        - first transition :
             <source> -> <midstate> <guard> <input> / <action>
        - second transition :
             <midstate> -> <target>  / <output>
    """
    trans_out=[]
    id=0
    for t in transitions:
        #midstate=t['source']+t['target'] + '_' + str(id)
        #print  t['output']
        if t['output'] == 'quiescence()':

          t['output']=''
          trans_out.append(t)
          continue

        midstate= 'm' + str(id)
        trans_out.append({
            'source': t['source'],
            'target': midstate,
            'guard' : t['guard'],
            'input' : t['input'],
            'output' : '',
            'action' : t['action']
        })
        trans_out.append({
            'source': midstate,
            'target': t['target'],
            'guard' : '',
            'input' : '',
            'output' : t['output'],
            'action' : ''
        })
        id=id+1
    return trans_out



def get_locations(transitions):
    """ Get set of all locations in statemachine """

    s=set()
    for t in transitions:
        s.add(t['source'])
        s.add(t['target'])

    return s

def get_uppaal_model(begin_state,transitions,interfaces,cons=[]):
    """Get uppaal model"""

    # TODO:
    # - make decl_str from cons,variables,interfaces :
    #     constants = 1,2;
    #     int v1, v2;
    #     void IIN2(int n2) {}
    # - add begin_state to model : <init ref="id11"/>


    declarations=[]
    cons=set(cons)
    #TODO: document more explicitly why -1 doesn't need to be added
    #cons.add(-1) # not necessary somehow??
    if cons: declarations.append('constants = ' + ','.join([str(x) for x in cons]) + ';')

     
    #if variables: declarations.append('int ' + ','.join(variables) + ';'

    # from interfaces make in uppaal :
    # - method declaration per input/output event
    # - two integer declarations for each parameter of an input
    #    `-> these integers are needed in applying mapper rules
    decs=[]
    for i in interfaces:
        param_types=[]
        pos=0
        for t in  i['param_types']:
            # create string for params in method declaration
            #   note: also fixes param type : Integer->int
            if t=='Integer':
                param_types.append('int p'+str(pos))
            else:
                param_types.append(t+' p'+str(pos))

            # for each INPUT Ixxx(p0,p1,...) add two integers for each of its parameters
            #  eg. IBLA(p0,p1)
            #     -> int first_IBLA_p0,last_IBLA_p0,first_IBLA_p1,last_IBLA_p1
            if i['method_name'][0] == 'I' :
                declarations.append('int first_' +  i['method_name'] + '_p' + str(pos) + ' = -1 ;')
                declarations.append('int last_' +  i['method_name'] + '_p' + str(pos) + ' = -1 ;')
            pos=pos+1
        decs.append( 'void ' + i['method_name'] + '(' + ','.join(param_types) + '){}' )

    declarations+=decs
    decl_str='\n'.join(declarations)

    system_str='Process = Template();system Process;'
    model= pyuppaal.NTA(declaration=decl_str,system=system_str)

    locs=[]
    initloc=None
    for loc in get_locations(transitions):
        upploc=pyuppaal.Location(id=loc,name=loc)
        locs.append( upploc )
        if loc==begin_state:
            initloc=upploc
 
 
    trans=[]
    for t in transitions:
        source=pyuppaal.Location(id=t['source'],name=t['source'])
        target=pyuppaal.Location(id=t['target'],name=t['target'])
        tran=pyuppaal.Transition(source, target, guard=t['guard'], synchronisation= t['input'] + t['output'] , assignment=t['action'])
 
        trans.append( tran )
 
 
    template=pyuppaal.Template(name='Template', locations=locs, initlocation=initloc, transitions=trans)
    model.add_template(template)
    return model




#-------------------------------------------------------------------------------
# old
#-------------------------------------------------------------------------------



def print_transitions_old(transitions,inputs,outputs,cons,number_params):
    indent='  '
    for t in transitions:
        print t['source'] + ' -> ' + t['target']

        inputlabel=t['input']['name']
        inputargs=t['input']['args']
        outputlabel=t['output']['name']
        outputargs=t['output']['args']

        # handle input and its arguments
        pos=0
        params=[]
        print indent, '_'.join( [inputlabel] + inputargs ) + ':'
        for i in inputargs:
            if int(i) != -1:
                rule=inputs[ inputlabel ][pos][int(i)]
                params.append( get_guard_from_input(rule,'==',cons,pos) )
            else:
                # all rules for inputarg gives != guard
                for rule in inputs[ inputlabel ][pos]:
                    params.append( get_guard_from_input(rule,'!=',cons,pos) )
            pos=pos+1

        num_input_params=number_params[inputlabel]


        inputparams=[]
        for num in xrange(num_input_params):
            inputparams.append('p'+str(num))

        print indent*2, 'input: ' + inputlabel + '(' + ','.join(inputparams)  + ')'
        if params:
            print indent*2,  'guard: ' + '[ '   + ' && '.join(params)  +  ' ]'
        else:
            print indent*2,  'guard: none'

        actions=[]
        for num in xrange(num_input_params):
            actions.append('first_' + inputlabel + '_p' + str(num) + ' = p' + str(num) + ';')
        action_first= '[ first_' +   inputlabel + '_p0 == -1 ] ' + ' '.join(actions)

        actions=[]
        for num in xrange(num_input_params):
            actions.append('last_' + inputlabel + '_p' + str(num) + ' = p' + str(num) + ';')
        action_last= '[ first_' +   inputlabel + '_p0 != -1 ] '  + ' '.join(actions)

        if num_input_params > 0 :
            print indent*2, 'duplicate in two transitions each with different guard and action :  '
            print indent*3,  '1. ' + action_first
            print indent*3,  '2. ' + action_last
        else :
            print indent*2, 'action: none'


        # handle output and its arguments
        pos=0
        params=[]
        print indent, '_'.join( [outputlabel] + outputargs )  + ':'
        for i in outputargs:
            if int(i) != -1:
                rule=outputs[ outputlabel ][pos][int(i)]
                params.append( get_output_param(rule,cons) )
            else:
                print "problem output  -1 !!!"

            pos=pos+1
        print indent*2, 'output: ' + outputlabel + '(' + ','.join(params) + ')'



def print_uppaal_template(nta,template=0):
    for t in nta.templates[template].transitions:
        print t.source.id + " -> " + t.target.id
        if t.guard.value: print   "   guard: " + t.guard.value
        if t.synchronisation.value: print  "   synchronisation: " + t.synchronisation.value
        if t.assignment.value: print  "   assignment: " + t.assignment.value

