"""Utilities for extended state machines.

get_lts_from_dotfile(dot_file) - Get label transition system from graphviz dot file
print_lts(start_state,transitions) - Print labeled transition system

parse_labels_of_lts(transitions)- Parse labels of labeled transition system
read_abstractions_from_jsonfile(json_filename) - Read data from learner
apply_mapper_data_on_transitions(..) - Apply data from mapper on learner results

print_transitions(transitions) - Print transitions
print_transitions_long(transitions) - Print transitions in long format
get_locations(transitions) - Get set of all locations in statemachine
split_input_and_output(transitions) - Split transitions with both an input
                                      and output event into two transitions
get_uppaal_model() - Get uppaal model


"""

# Author: Harco Kuppens

__all__ = ["get_lts_from_dotfile",
           "print_lts",
           "parse_labels_of_lts",
           "read_abstractions_from_jsonfile",
           "apply_mapper_data_on_transitions",
           "print_transitions",
           "print_transitions_long",
           "get_locations",
           "split_input_and_output",
           "get_uppaal_model",
          ]

import sys, re, json, copy # module from standard lib (python 2.6 and later)

import pyuppaal
#import pygraphviz

import pprint

#-------------------------------------------------------------------------------
#  pre processing
#-------------------------------------------------------------------------------



def get_lts_from_dotfile(dot_file):
    """ Get labeled transition system from graphviz dot file

        The dot file:
           - describes a digraph with labels
           - encodes the start state with the color='red' attribute
             note: this corresponds with the highlighted state in learnlib API
    

        Returns: [start_state,transions]
        Where :
           - start_state: start state label
           - transitions: list of transitions
        
        Note: when parsing dot file it throws away transitions with the keywords : 
        
                 quiescence or inconsistency or undefined

    """

    start_state='unknown'
    f=file(dot_file)
    lines=f.readlines()

    # find start state
    for line in lines:
        line=line.replace('label=""','')
        if line.find('[color="red"]') != -1:
            start_state=line[:line.find(' ')]
            break


    # get transitions
    transitions=[]
    for line in lines:
        if line.find('->') != -1:
            transitions.append(line)

    # throw away transitions with the keywords : quiescence or inconsistency or undefined
    transitions = [ t for t in transitions if ( 'quiescence' not in t ) and ( 'inconsistency' not in t  )  and ( 'undefined' not in t  )]

    trans_out=[]
    regexpr_transition=re.compile(r'\s*(\w*)\s*-\>\s*(\w*)\s*\[label=\<(.*)\>\]')
    regexpr_tag=re.compile(r'<[^>]+>')
    for transition in transitions:
        match=regexpr_transition.match(transition)
        if match:
            match=match.groups()
            label=regexpr_tag.sub('',match[2])
            trans_out.append({
              'source' : match[0],
              'target' : match[1],
              'label': label
            })

    return [start_state,trans_out]

#     trans_out=[]
#     start_state='unknown'
#     G=pygraphviz.AGraph(dot_file)
#     transitions=G.edges()
#     for t in transitions:
#         label=t.attr['label']
#         label=re.sub('<[^<]+?>', '', label)
#         trans_out.append({
#           'source' : str(t[0]),
#           'target' : str(t[1]),
#           'label': label
#         })
#     for n in G.nodes():
#         if n.attr['color'] == 'red':
#             start_state=n.attr['label']
#     return [start_state,trans_out]

def print_lts(start_state,transitions):
    """Print labeled transition system

       The transitions are printed as used in a statemachine::

          source -> target : label

    """
    indent='  '
    print "start state: " + start_state
    for t in transitions:
        print t['source'] + ' -> ' + t['target']  + ' : ' + t['label']

def parse_abstract_input(inputstr):
        inputparts=inputstr.split('_')
        inputlabel=inputparts[0]
        inputargs=inputparts[1:]
        return (inputlabel,inputargs)

def parse_abstract_output(outputstr):
        outputupdates={}        
        outputlabel="parseOutputError"
        output=outputstr.split('_')
        prefixLength=0
        for piece in output:  
            if '=' in piece:
                (statevar,update)=piece.split('=')
                outputupdates[statevar]=update
                prefixLength=prefixLength+1
            elif piece.startswith("x"):
                statevar=piece
                update="-1" # meaning unset statevar
                if not outputupdates.has_key(statevar): 
                  outputupdates[statevar]=update 
                prefixLength=prefixLength+1                                               
            else:
                outputlabel=piece
                break 
        # slice updates and label from output list -> output args
        outputargs=output[prefixLength+1:]         
        # note: we assume valid outputargs 
        return (outputupdates,outputlabel,outputargs)

def parse_labels_of_lts(transitions):
    """Parse labels of labeled transition system
    """
    trans_out=[]
    for t in transitions:
        
        label=t['label']
        [inputstr,outputstr]=label.split('/')
        (inputlabel,inputargs)=parse_abstract_input(inputstr)
        (outputupdates,outputlabel,outputargs)= parse_abstract_output(outputstr)
        parsed_t={
          'source' : t['source'],
          'target' : t['target'],
          'input': {
              'name': inputlabel,
              'args': inputargs
          },
          'output': {
              'name': outputlabel,
              'args': outputargs,
              'updates': outputupdates
          },
          'abstractinput':  inputstr,
          'abstractoutput': outputstr,
          'altValues': None,
        }
        trans_out.append(parsed_t)

    stateXmethodY2translist={}
    for t in  trans_out:
         stateXmethodY=(t['source'],t['input']['name'])
         if stateXmethodY in  stateXmethodY2translist:
             translist=stateXmethodY2translist[stateXmethodY]
         else:
             translist=[]
             stateXmethodY2translist[stateXmethodY]=translist
         translist.append(t)

    for translist in stateXmethodY2translist.values():
         altValues={}
         for paramIndex in range(len(translist[0]['input']['args'])):
             altValues[paramIndex]=[]
             alternative_transition = None      # for current paramIndex!!
             for transition in translist:
                 abstractValue=transition['input']['args'][paramIndex]
                 if int(abstractValue) == -1:
                    transition['altValues']=altValues
                 else:
                    altValues[paramIndex].append( int(abstractValue)  )

    return trans_out




def read_abstractions_from_jsonfile(json_filename_in):
    """ Read data from learner

        Reads in json file which contains data from learned model
        learned with learner.
    """
    f=open(json_filename_in,'r')
    mapper=json.load( f )
    f.close()
    #print json.dumps(jsonstr, sort_keys=True, indent=4)


    # output abstactions : no output abstraction in json file for statevar mode : output abstractions are in dot file!
    output_abstractions=None
 
    # input abstactions
    input_abstractions={}
    ins=mapper['inputs']
    for i in ins:
        input_abstractions[ i['name'] ] = i['params']
 
 
    # sut interfaces
    interfaces=mapper['concrete_alphabet']


    cons=mapper['constants']


#         {
#             "name": "IRegister", 
#             "params": [
#                 [], 
#                 [
#                     {
#                         "statevarindex": 0, 
#                         "statevartype": "p"
#                     }
#                 ]
#             ]
#         }


    return  [input_abstractions,output_abstractions,cons,interfaces]



#-------------------------------------------------------------------------------
#  processing
#-------------------------------------------------------------------------------
# abstraction in  firstlast and statemode
#     {
#         "methodname": "IRegister", 
#         "paramindex": 1, 
#         "stateid": 0
#     }
# 

# input abstraction in statevar mode
#     {
#         "statevarindex": 0, 
#         "statevartype": "p"
#     }
# note: output abstractions are generated on the fly from memV and are stored in dot model
def get_guard_from_input(rule,operator,cons,pos,inputlabel,source_state):
    if rule['statevartype'] == "c":
        return "p" + str(pos) + ' ' + operator + ' ' +  str( cons[rule['statevarindex']] )
    elif rule['statevartype'] == "x":
        return "p" + str(pos) + ' ' + operator + ' x' +  str( rule['statevarindex'])    
    elif rule['statevartype'] == "p":
        return "p" + str(pos) + ' ' + operator + ' p' +  str( rule['statevarindex'] )
        
        
#     if rule['methodname'] == '_CONSTANTS_':
#         return "p" + str(pos) + ' ' + operator + ' ' +  str( cons[rule['paramindex']] )
#     else:
#         stateid='s' + str(rule['stateid'])
#         if  ( rule['methodname'] == inputlabel ) and ( stateid == source_state and pos > rule['paramindex'] ) :  
#             # backrefs always point back to earlier params in trace
#             # note: that a selfloop transition can be called twice in which p0 of
#             #       the second call points back to p0 of the first call!!
#             return "p" + str(pos) + ' ' + operator + ' ' + 'p' +  str(rule['paramindex'])
#         else:
#             return "p" + str(pos) + ' ' + operator + ' s' + str(rule['stateid']) + '_' + rule['methodname']  + '_p' +  str(rule['paramindex'])


def get_input_event(inputlabel,num_input_params):
        # construct input event and its arguments for transition
        # -> constructed from : inputlabel and num_input_params
        inputparams=[]
        for num in xrange(num_input_params):
            inputparams.append('p'+str(num))
        return inputlabel + '(' + ','.join(inputparams)  + ')'



def get_guard(altValues,inputlabel,inputargs,cons,input_abstractions,source_state):
        # construct guard for transition
        # -> constructed from :  inputlabel,inputargs,cons  and input_abstractions(rules)
        # 1. guards are retreived by applying input_rules on input
        parameter_index=0
        guards=[]
        for abstraction_index in inputargs:
            if int(abstraction_index) != -1: 
                rule=input_abstractions[ inputlabel ][parameter_index][int(abstraction_index)]
                guards.append( get_guard_from_input(rule,'==',cons,parameter_index,inputlabel,source_state) )
                
                if not altValues == None: 
                  for anti_abstraction_index in altValues[parameter_index]:
                      if  int(anti_abstraction_index) < int(abstraction_index):
                          rule=input_abstractions[ inputlabel ][parameter_index][int(anti_abstraction_index)]
                          guards.append( get_guard_from_input(rule,'!=',cons,parameter_index, inputlabel,source_state ) )                
            else:
                # all alt. rules for inputarg gives != guard
                #for rule in input_abstractions[ inputlabel ][parameter_index]:
                #print altValues,parameter_index

                for anti_abstraction_index in altValues[parameter_index]:
                    rule=input_abstractions[ inputlabel ][parameter_index][int(anti_abstraction_index)]
                    guards.append( get_guard_from_input(rule,'!=',cons,parameter_index, inputlabel,source_state ) )
            parameter_index=parameter_index+1
        # 2. store collect guard conditions joined with boolean AND in transition
        if guards:
            return ' && '.join(guards)
        else:
            return '' # no guard

def is_local_var_needed(statevar, outputargs, outputupdates):
       
    # 1. we skip self assignments (eg. x0=x0)
    #    `-> no local variable needed then 
    if outputupdates.has_key(statevar) and outputupdates[statevar] == statevar:  
        return False
       
    # 2. check statevar is used in output action
    if (statevar in outputupdates) and (statevar in outputargs) :
        # print "LOCAL NEEDED statevar", statevar, "in outputargs",  outputargs 
        return True 

    # 3. check statevar is used as value in update action (because in update action we need to use the original value)
    #  -> local variable needed to garantee the parallel update of the prefix of the output abstraction!
    if (statevar in outputupdates) and (statevar in  outputupdates.values()):
        # print "LOCAL NEEDED statevar", statevar, "in outputupdates", outputupdates 
        return True 
    
    return False

def get_update_actions_from_outputlabel_only(statevars,outputargs,outputupdates,cons):
    actions=[]
    # 1. if local variable is needed for statevar
    #       - add local statevar
    #       - add action to assign statevar to local statevar before doing updates
    #         e,g, local_x2=x2  
    for (statevar,update) in outputupdates.items():
        #print "update: ",statevar, " = " , update
        if is_local_var_needed(statevar,outputargs,outputupdates):
            #print "---> local var needed "
            local_statevar=  "local_" + statevar 
            statevars.add(local_statevar)
            keep_old_in_local_statevar_action= local_statevar + "=" + statevar + ";"
            #print "keep_old_in_local_statevar_action : ", keep_old_in_local_statevar_action
            actions.append(keep_old_in_local_statevar_action)
    # 2. get all actions from outputupdates
    for (statevar,update) in outputupdates.items():
        statevars.add(statevar)
        
        # skip assignment when assigning  statevar to itself e.g. x0=x0   
        if statevar == update:
            continue
                  
        # add action to assign new value to statevar
        if update.startswith("p"):
            action= statevar + "=" + update + ";"
        elif update.startswith("x"):
            if is_local_var_needed(update,outputargs,outputupdates):
                local_statevar=  "local_" + update
                action= statevar + "=" + local_statevar + ";"
                statevars.add(local_statevar)
            else:
                action= statevar + "=" +  update + ";"
        elif str(update) == "-1":  # meaning unset statevar
            #action=""
            action= statevar + "=-1000000;"                         
            #action= statevar + "=-1;"                         
        elif update.startswith("c"):
            consIndex=int(update[1:])
            value=cons[consIndex]
            action= statevar + "=" +  str(value)  + ";"                
        else:    
            print "error: update doesn't start with x,c nor p!!  update: ", update
            sys.exit(1)
        actions.append(action)
    # 3. for each pX in an output we :
    #     - add statevar 'local_pX'
    #     - add for corresponding input transition the action 'local_pX=pX'
    #     - and rename pX in output to 'local_pX' 
    #            `-> we do this in method 'get_output_event_using_outputlabel_only'
    for indx,outputarg in enumerate(outputargs):
        if outputarg.startswith("p"):
            statevar="local_" + outputarg
            statevars.add(statevar)
            action= statevar + "=" + outputarg + ";"
            actions.append(action)
    # 4. combine all actions in single result action
    if actions:
        return ' '.join(actions)
    else:
        return '' # no action

def get_output_event_using_outputlabel_only(statevars,outputlabel,outputargs,outputupdates,cons):
    params=[]
    #print outputargs

        
    for arg in outputargs:
        param=arg
        if arg.startswith("x") and  is_local_var_needed(param,outputargs,outputupdates):
            param="local_" + param
            statevars.add(param)
        if arg.startswith("p"):
            param="local_" + param
            statevars.add(param)
        if arg.startswith("c"):
            consIndex=int(arg[1:])
            param=str(cons[consIndex])
        params.append(param)
    #print outputlabel + '(' + ','.join(params) + ')'
    return outputlabel + '(' + ','.join(params) + ')'



def apply_mapper_data_on_transitions(transitions,input_abstractions,output_abstractions,cons,number_params,giveAbstractInputsInstead):
    """ Apply data from mapper on results from learner

   """

    

    # an input abstaction i for some method param is of the form : Sx_In_Pj which specifies a storage point (=action parameter index)
    # loop over all input abstractions and fetch :
    #   - statevars: statevar for each unique storage point Sx_In_Pj
    #   - storage_index: map each state's input method 'Sx_In' to a set of storage points 'Sx_In_Pj'  which store date for that state's input
    abstractions = input_abstractions.values() 

# [[], [[], [{u'statevarindex': 0, u'statevartype': u'p'}]], [[], []]]

    #abstractions = input_abstractions.values() + output_abstractions.values()   # old version used same input as output abstractions using Sx_In_Pj storage points (=action parameter index)
    statevars=set()
    for  action_abstractions in  abstractions:  
      # abstractions for some action X
      for  param_abstractions in  action_abstractions:
        # abstractions for some action X and parameter i of that action : per action parameter we can have multiple abstractions!
        for abstraction in  param_abstractions:
            if abstraction:
                abstraction['statevartype'].startswith("x")
                statevar =  "x" + str(abstraction['statevarindex'])
                statevars.add(statevar)

    trans_out=[]
    indent='  '
    abstractinputs=set()
    for t in transitions:
        #print t
        transition={}
        source=transition['source']=t['source']   # eg. s0  from dot file
        target=transition['target']=t['target']


        inputlabel=t['input']['name']
        inputargs=t['input']['args']
        outputlabel=t['output']['name']
        outputargs=t['output']['args']
        outputupdates=t['output']['updates']
        #print outputlabel,outputargs,outputupdates

        #transition['abstract']=t
        altValues=t['altValues']
        
        num_input_params=number_params[inputlabel]


        if giveAbstractInputsInstead:
            if inputargs:
                abstractinput="O"+inputlabel + "_"  + "_".join(inputargs)
            else:
                # inputs without params cannot have abstractions, and therefore have no subindex
                abstractinput="O"+inputlabel 
               
            abstractinput=abstractinput.replace('-','min')
            transition['output']=abstractinput +"()"
            abstractinputs.add( abstractinput )
        else:
            transition['output']=get_output_event_using_outputlabel_only(statevars,outputlabel,outputargs,outputupdates,cons)

        #transition['output']=get_output_event(outputlabel,outputargs,cons,output_abstractions)

        transition['input']=get_input_event(inputlabel,num_input_params)
        transition['guard']=get_guard(altValues,inputlabel,inputargs,cons,input_abstractions,transition['source'])

        #actionstr = get_actions_to_store_variables_for_input_abstractions(storage_index,inputlabel,transition['source'])  + get_update_actions_from_outputlabel_only(statevars,outputargs,outputupdates,cons)
        #TODO
        actionstr = get_update_actions_from_outputlabel_only(statevars,outputargs,outputupdates,cons)

        transition['action'] = actionstr
        #get_actions(inputlabel,inputargs,cons,inputs,transition['source'])
        trans_out.append(transition)
        

    
    return (trans_out,statevars,abstractinputs)


#-------------------------------------------------------------------------------
# post processing
#-------------------------------------------------------------------------------



def print_transitions_long(transitions):
    """Print transitions in long format

       For each transition the guard,input,output,action are printed
       on a separate line::

          <source> -> <target>
             guard: <guard>
             input: <input>
             output: <output>
             action: <action>

    """
    indent='  '
    for t in transitions:
        print t['source'] + ' -> ' + t['target']
        print indent*2, 'guard:',  t['guard']
        #print indent*2, 'guard:', '[ ', t['guard'] , ' ]'
        print indent*2, 'input:', t['input']
        print indent*2, 'output:', t['output']
        print indent*2, 'action:', t['action']

def print_transitions(transitions):
    """Print transitions

       The transitions are printed as used in a statemachine::

          source -> target : [guard] input/output ; action

    """
    indent='  '
    for t in transitions:
        guard=''
        if  t['guard']:
            guard= " [ " + t['guard']  + ' ] '
        action=''
        if  t['action']:
            action = " ; " + t['action']
        print t['source'] + ' -> ' + t['target']  + ' : ' + guard +  t['input'] + ' / ' +  t['output'] +  action
 
def split_input_and_output(transitions):
    """Split transitions with both an input and output event into two transitions

       Makes two sequential transitions with a dummy state in between:
        - dummy state <midstate> is labeled :
             m_<counter>
        - first transition :
             <source> -> <midstate> <guard> <input> / <action>
        - second transition :
             <midstate> -> <target>  / <output>
    """
    trans_out=[]
    id=0
    for t in transitions:
#        # if output is quiescence then do not split transition, but just keep it as input transition (with empty output)
#        if t['output'] == 'quiescence()':
#
#          t['output']=''
#          trans_out.append(t)
#          #forget?? : id=id+1
#          continue

        midstate= 'm' + str(id)
        trans_out.append({
            'source': t['source'],
            'target': midstate,
            'guard' : t['guard'],
            'input' : t['input'],
            'output' : '',
            'action' : t['action']
        })
        trans_out.append({
            'source': midstate,
            'target': t['target'],
            'guard' : '',
            'input' : '',
            'output' : t['output'],
            'action' : ''
        })
        id=id+1
    return trans_out



def get_locations(transitions):
    """ Get set of all locations in statemachine """

    s=set()
    for t in transitions:
        s.add(t['source'])
        s.add(t['target'])

    return s

def get_uppaal_model(begin_state,transitions,interfaces,statevars,giveAbstractInputsInstead,abstractinputs,cons=[]):
    """Get uppaal model"""

    # TODO:
    # - make decl_str from cons,variables,interfaces :
    #     constants = 1,2;
    #     int v1, v2;
    #     void IIN2(int n2) {}
    # - add begin_state to model : <init ref="id11"/>


    declarations=[]
    cons=set(cons)
    #if cons: declarations.append('constants = ' + ','.join([str(x) for x in cons]) + ';')
    for c in cons:
        declarations.append('const int CONSTANT' + str(c) + ' = ' + str(c) + ';')
       


     
    #if variables: declarations.append('int ' + ','.join(variables) + ';'

    # from interfaces make in uppaal :
    # - method declaration per input/output event
    # - two integer declarations for each parameter of an input
    #    `-> these integers are needed in applying mapper rules
    decs=[]
    if giveAbstractInputsInstead:
        for i in abstractinputs:
          decs.append( 'void ' + i + '(){}' )

    for i in interfaces:
        param_types=[]
        pos=0

        for t in  i['param_types']:
            # create string for params in method declaration
            #   note: also fixes param type : Integer->int
            if t=='Integer':
                param_types.append('int p'+str(pos))
            else:
                param_types.append(t+' p'+str(pos))

#             # for each INPUT Ixxx(p0,p1,...) add two integers for each of its parameters
#             #  eg. IBLA(p0,p1)
#             #     -> int first_IBLA_p0,last_IBLA_p0,first_IBLA_p1,last_IBLA_p1
#             if i['method_name'][0] == 'I' :
#                 declarations.append('int first_' +  i['method_name'] + '_p' + str(pos) + ' = -1 ;')
#                 declarations.append('int last_' +  i['method_name'] + '_p' + str(pos) + ' = -1 ;')
            pos=pos+1

        # only add input declarations when giving abstractInputs as output instead
        if ( not giveAbstractInputsInstead ) or i['method_name'][0] == "I" :
            decs.append( 'void ' + i['method_name'] + '(' + ','.join(param_types) + '){}' )

    for statevar in statevars:
        declarations.append( "int " +statevar +" = -1 ;")

    declarations+=decs
    decl_str='\n'.join(declarations)

    system_str='Process = Template();system Process;'
    model= pyuppaal.NTA(declaration=decl_str,system=system_str)

    locs=[]
    initloc=None
    for loc in get_locations(transitions):
        upploc=pyuppaal.Location(id=loc,name=loc)
        locs.append( upploc )
        if loc==begin_state:
            initloc=upploc
 
 
    trans=[]
    for t in transitions:
        source=pyuppaal.Location(id=t['source'],name=t['source'])
        target=pyuppaal.Location(id=t['target'],name=t['target'])
        tran=pyuppaal.Transition(source, target, guard=t['guard'], synchronisation= t['input'] + t['output'] , assignment=t['action'])
 
        trans.append( tran )
 
 
    template=pyuppaal.Template(name='Template', locations=locs, initlocation=initloc, transitions=trans)
    model.add_template(template)
    return model




