#!/usr/bin/env python
"""
 run tomte learner 

 usage:  tomte_learn  config.yaml

   where config.yaml  specifies :
      - learning parameters
      - sutinfo file which describes alphabet used by sut 
      - how to connect to sut (default is using socket localhost:7892

"""
# Author: Harco Kuppens


import sys, re, pprint,os  # modules from standard lib (python 2.6 and later)

import tempfile
import shutil
import argparse
import yaml

import util
import learnresult2uppaal


def check_dependencies():
    util.check_program('java','.exe')

def extend_short_usage_msg_with_help_info_msg(argparser):
    usage=argparser.print_usage
    def newusage(obj):
       usage()
       print "Try `%s --help' for more information." % argparser.prog
    argparser.print_usage=newusage

def args_parsing(args):

    argparser = argparse.ArgumentParser(prog=os.path.basename(args[0]))

    #argparser.add_argument("--max-traces", action='store'  ,help='Maximum number of traces to run during equivalence testing.')
    #argparser.add_argument("--min-trace-length", action='store', help="Minimum length of traces during equivalence query.")
    #argparser.add_argument("--max-trace-length", action='store',help="Maximum length of traces during equivalence query.")

    argparser.add_argument("--seed", action='store', help="Seed to use for random number generator.")	
    argparser.add_argument("--output-dir", action='store',help="Directory to store output.")
    argparser.add_argument('--port',action='store',help='tcp port  to listen on for incoming connections')
    argparser.add_argument("-e","--eclipse", action='store_true',help="Run using build code from eclipse project.")
    argparser.add_argument('--config-option',action='append', help="Overrule a configuration option in config.yaml file.")



    argparser.add_argument('configfile', help='learner configuration file')

    extend_short_usage_msg_with_help_info_msg(argparser)

    parse_result=argparser.parse_args(args[1:])


    seed=parse_result.seed
    output_dir=parse_result.output_dir
    port=parse_result.port
    eclipse=parse_result.eclipse
    config_options=parse_result.config_option

    # handle input/output file argument
    configfile=os.path.abspath(parse_result.configfile).replace('\\','/')
    util.check_isreadablefile(configfile)

    return  (configfile,output_dir,seed,port,eclipse,config_options)

def get_tomte_external_libraries():
    tomte_external_libraries=[
        "lib/learnlib-distribution-20110714.jar",
        "lib/api-0.4.7.jar",
        "lib/json_simple-1.1.jar",
        "lib/snakeyaml-1.10.jar",
        "lib/log4j-1.2.17.jar",
        "lib/gson-2.2.4.jar",
        "lib/jbeandumper.1.5.1.jar",
        "lib/junit-4.10.jar",
    ]
    return tomte_external_libraries

def get_tomte_full_classpath(tomte_root_dir,eclipse):
    if eclipse:
        tomte_classpath="build/"
        tomte_classpath=os.path.abspath(os.path.join(tomte_root_dir,tomte_classpath))
    else:
        tomte_classpath="lib/tomte.jar"
        tomte_classpath=os.path.abspath(os.path.join(tomte_root_dir,tomte_classpath))
        util.check_isreadablefile(tomte_classpath)

    tomte_external_libraries=get_tomte_external_libraries()
    tomte_external_libraries = [ os.path.abspath(os.path.join(tomte_root_dir,entry)) for entry in tomte_external_libraries ]
    for path in tomte_external_libraries:
       util.check_isreadablefile(path)
    classpath = [tomte_classpath] + tomte_external_libraries
    classpath = os.path.pathsep.join(classpath)
    return classpath

def run_java_program(classpath,mainclass,java_program_args):
    command=[ 'java', '-cp', classpath, mainclass] + java_program_args
    print("running command : \n   "  + util.join_command(command) )
    run_command(command)

def run_command(command):
    """
      run command
      if command exits with error
         than this function will also exit the current program with the error
      else
         just returns from this function
    """
    import subprocess
    try:
      process = subprocess.Popen(command)
    except Exception as inst:
      print("problem running command : \n   "  + util.join_command(command) + "\n problem : " + str(inst))

    process.communicate()  # no pipes set for stdin/stdout/stdout streams so does effectively only just wait for process ends  (same as process.wait()

    if process.returncode:
       print "problem:  command exited with errorcode " + str(process.returncode)
       sys.exit(process.returncode)


def learn(configfile,output_dir,seed,port,tomte_root_dir,tomte_classpath,config_options):
    """
	 run tomte learner 

	 usage:  tomte_learn  config.yaml

	   where config.yaml  specifies :
	      - learning parameters
	      - sutinfo file which describes alphabet used by sut 
              - how to connect to sut (default is using socket localhost:7892

    """

    # read config and sutinfo files to retreive sut's name
    configdir=os.path.dirname(configfile)
   

    configdata=yaml.load(file(configfile))

    # check first if sutinfoFile is really set in config.yaml file 
    if not configdata.has_key('learning'): 
        print "you must specify learning configuration in configuration"
        sys.exit(1)

    if not ( configdata['learning'].has_key('useSutSimulation') and configdata['learning']['useSutSimulation'] ):
        if not configdata['learning'].has_key('sutinfoFile'):
            print "you must specify sutinfoFile in configuration"
            sys.exit(1)

        # we take path to sutinfoFile to be relative from the directory where the config.yaml is located (configDir)
        # TODO: also allow absolute dir for sutinfoFile
        sutinfoFile=os.path.normpath( os.path.join(configdir, configdata['learning']['sutinfoFile']) )
        if not util.isreadablefile(sutinfoFile):
            print "\nPROBLEM:\n   cannot read the file : " + sutinfoFile +"\n\n"
            sys.exit(1)

        #sutinfoFile=configdata['learning']['sutinfoFile']

        sutinfoData=yaml.load(file(sutinfoFile))


        sutname=sutinfoData['name']
    else:
        #sutname="unknown_sutname" # should be dirname!
        sutname=os.path.basename(configdir)

    import time
    timestamp=time.strftime("%y-%m-%d_%H.%M.%S")


    # set parameterized output dir here so that after learning we know where 
    # to access the results (MUST now timetamp!!) 
    if output_dir:
       outputDir=output_dir
    else:
       if configdata.has_key('learnResults') and configdata['learnResults'].has_key('outputDir'):
          outputDir=configdata['learnResults']['outputDir']
       else:
          outputDir="output/${timestamp}_${sutname}"
    outputDir=outputDir.replace("${timestamp}",timestamp).replace("${sutname}",sutname)




    # 1. run tomte java application
    #-------------------------------

    mainclass= 'abslearning.learner.Main'
    classpath=get_tomte_full_classpath(tomte_root_dir,tomte_classpath)

    # tomte's java command args
    java_program_args=[]
    #
    # required options:
    java_program_args+=['--tomte-root-path',tomte_root_dir]
    java_program_args+=['--output-dir',str(outputDir)]
    # optionally options:
    if seed: java_program_args+=['--seed',str(seed)]
    if port: java_program_args+=['--port',str(port)]
    if config_options:
        for config_option in config_options:
            java_program_args+=['--config-option',config_option]
    # normal arguments
    java_program_args.append(configfile)

    run_java_program(classpath,mainclass,java_program_args)



    print 'learning done'

    # 2. make uppaal model from learnresults 
    #---------------------------------------

    # read dumped config file by learner -> reason : in this file all defaults are explicit set

    configdata=yaml.load(file(outputDir + '/input/config.effective.yaml'))

    json_filename_in = outputDir + '/' + configdata['learnResults']['learnedModelDataFile']
    dot_filename_in  = outputDir + '/' + configdata['learnResults']['abstractModelDotFile']
    model_file_out   = outputDir + '/' + configdata['learnResults']['learnedConcreteModelFile']


#     learnedModelDataFile="learnedModelData.json"
#     abstractModelDotFile="learnedAbstractModel.dot"
#     learnedConcreteModelFile="learnedConcreteModel.xml"
#
#     if configdata.has_key('learnResults'):
#         if configdata['learnResults'].has_key('learnedModelDataFile'):
#            learnedModelDataFile=configdata['learnResults']['learnedModelDataFile']
#         if configdata['learnResults'].has_key('abstractModelDotFile'):
#            abstractModelDotFile=configdata['learnResults']['abstractModelDotFile']
#         if configdata['learnResults'].has_key('learnedConcreteModelFile'):
#            learnedConcreteModelFile=configdata['learnResults']['learnedConcreteModelFile']
#
#    json_filename_in = outputDir + '/' + learnedModelDataFile
#    dot_filename_in  = outputDir + '/' + abstractModelDotFile
#    model_file_out   = outputDir + '/' + learnedConcreteModelFile


#    learnresult2uppaal.learnresult2uppaal(json_filename_in,dot_filename_in,model_file_out)


    print "\n\n"
    print "results are stored in directory : " + outputDir 
    print "learned concrete model: " +  model_file_out 
    print "\n\n"





def main(args):
    """
	 run tomte learner 

	 usage:  tomte_learn  config.yaml

	   where config.yaml  specifies :
	      - learning parameters
	      - sutinfo file which describes alphabet used by sut 
              - how to connect to sut (default is using socket localhost:7892


    """
    check_dependencies()

    (configfile,output_dir,seed,port,eclipse,config_options)=args_parsing(args)
    current_script_dir=os.path.dirname(sys.argv[0])
    tomte_root_dir= os.path.normpath(os.path.join(current_script_dir,'..') )
    learn(configfile,output_dir,seed,port,tomte_root_dir,eclipse,config_options)



if __name__ == "__main__":
    print main(sys.argv)

